@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2011 Douglas Gilbert.
+ * Copyright (c) 2006-2014 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,14 +42,23 @@
 #include <ctype.h>
 #include <getopt.h>
 
+
 #include "sg_lib.h"
+
+#ifndef _WIN32_WINNT
+#define _WIN32_WINNT 0x0602
+/* claim its W8 */
+#endif
+
 #include "sg_pt_win32.h"
 
+#include "ddpt.h"
+
 /*
  * This is called when the '--wscan' option is given to ddpt. It is
  * Win32 only code and shows the relationship between various device names
- * and volumes in Windows OSes (Windows 2000, 2003, XP, Vista and W7). There
- * is an optional scsi adapter scan.
+ * and volumes in Windows OSes (Windows 2000, 2003, XP, Vista, W7 and W8).
+ * There is an optional scsi adapter scan.
  */
 
 #define MAX_SCSI_ELEMS 1024
@@ -71,6 +80,7 @@
     CTL_CODE(IOCTL_STORAGE_BASE, 0x0500, METHOD_BUFFERED, FILE_ANY_ACCESS)
 
 
+#ifndef _DEVIOCTL_
 typedef enum _STORAGE_BUS_TYPE {
     BusTypeUnknown      = 0x00,
     BusTypeScsi         = 0x01,
@@ -86,7 +96,9 @@ typedef enum _STORAGE_BUS_TYPE {
     BusTypeSata         = 0x0B,
     BusTypeSd           = 0x0C,
     BusTypeMmc          = 0x0D,
-    BusTypeMax          = 0x0E,
+    BusTypeVirtual             = 0xE,
+    BusTypeFileBackedVirtual   = 0xF,
+    BusTypeMax,
     BusTypeMaxReserved  = 0x7F
 } STORAGE_BUS_TYPE, *PSTORAGE_BUS_TYPE;
 
@@ -105,6 +117,7 @@ typedef struct _STORAGE_DEVICE_DESCRIPTOR {
     ULONG RawPropertiesLength;
     UCHAR RawDeviceProperties[1];
 } STORAGE_DEVICE_DESCRIPTOR, *PSTORAGE_DEVICE_DESCRIPTOR;
+#endif
 
 typedef struct _STORAGE_DEVICE_UNIQUE_IDENTIFIER {
     ULONG  Version;
@@ -117,6 +130,7 @@ typedef struct _STORAGE_DEVICE_UNIQUE_IDENTIFIER {
 // Use CompareStorageDuids(PSTORAGE_DEVICE_UNIQUE_IDENTIFIER duid1, duid2)
 // to test for equality
 
+#ifndef _DEVIOCTL_
 typedef enum _STORAGE_QUERY_TYPE {
     PropertyStandardQuery = 0,
     PropertyExistsQuery,
@@ -139,6 +153,7 @@ typedef struct _STORAGE_PROPERTY_QUERY {
     STORAGE_QUERY_TYPE QueryType;
     UCHAR AdditionalParameters[1];
 } STORAGE_PROPERTY_QUERY, *PSTORAGE_PROPERTY_QUERY;
+#endif
 
 
 /////////////////////////////////////////////////////////////////////////////
@@ -233,6 +248,10 @@ get_bus_type(int bt)
         return "Sd   ";
     case BusTypeMmc:
         return "Mmc  ";
+    case BusTypeVirtual:
+        return "Virt ";
+    case BusTypeFileBackedVirtual:
+        return "FBVir";
     case BusTypeMax:
         return "Max  ";
     default:
@@ -246,7 +265,8 @@ query_dev_property(HANDLE hdevice,
 {
     DWORD num_out, err;
     char b[256];
-    STORAGE_PROPERTY_QUERY query = {StorageDeviceProperty, PropertyStandardQuery, {0} };
+    STORAGE_PROPERTY_QUERY query = {StorageDeviceProperty,
+                                    PropertyStandardQuery, {0} };
 
     memset(data, 0, sizeof(*data));
     if (! DeviceIoControl(hdevice, IOCTL_STORAGE_QUERY_PROPERTY,
@@ -254,14 +274,14 @@ query_dev_property(HANDLE hdevice,
                           &num_out, NULL)) {
         if (verbose > 2) {
             err = GetLastError();
-            fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(Devprop) failed, "
+            pr2serr("  IOCTL_STORAGE_QUERY_PROPERTY(Devprop) failed, "
                     "Error=%ld %s\n", err, get_err_str(err, sizeof(b), b));
         }
         return -ENOSYS;
     }
 
     if (verbose > 3)
-        fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevProp) num_out=%ld\n",
+        pr2serr("  IOCTL_STORAGE_QUERY_PROPERTY(DevProp) num_out=%ld\n",
                 num_out);
     return 0;
 }
@@ -281,13 +301,13 @@ query_dev_uid(HANDLE hdevice,
                           &num_out, NULL)) {
         if (verbose > 2) {
             err = GetLastError();
-            fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) failed, "
+            pr2serr("  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) failed, "
                     "Error=%ld %s\n", err, get_err_str(err, sizeof(b), b));
         }
         return -ENOSYS;
     }
     if (verbose > 3)
-        fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) num_out=%ld\n",
+        pr2serr("  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) num_out=%ld\n",
                 num_out);
     return 0;
 }
@@ -341,8 +361,8 @@ enum_scsi_adapters(void)
         if (fh != INVALID_HANDLE_VALUE) {
             hole_count = 0;
             success = DeviceIoControl(fh, IOCTL_SCSI_GET_INQUIRY_DATA,
-                                      NULL, 0, inqDataBuff, sizeof(inqDataBuff),
-                                      &dummy, FALSE);
+                                      NULL, 0, inqDataBuff,
+                                      sizeof(inqDataBuff), &dummy, FALSE);
             if (success) {
                 PSCSI_BUS_DATA pbd;
                 PSCSI_INQUIRY_DATA pid;
@@ -363,8 +383,10 @@ enum_scsi_adapters(void)
                                  pid->PathId, pid->TargetId, pid->Lun);
                         printf("%-15s", b);
                         snprintf(b, sizeof(b) - 1, "claimed=%d pdt=%xh %s ",
-                                 pid->DeviceClaimed, pid->InquiryData[0] % 0x3f,
-                                 ((0 == pid->InquiryData[4]) ? "dubious" : ""));
+                                 pid->DeviceClaimed,
+                                 pid->InquiryData[0] % 0x3f,
+                                 ((0 == pid->InquiryData[4]) ? "dubious" :
+                                                                ""));
                         printf("%-26s", b);
                         printf("%.8s  %.16s  %.4s\n", pid->InquiryData + 8,
                                pid->InquiryData + 16, pid->InquiryData + 32);
@@ -373,16 +395,16 @@ enum_scsi_adapters(void)
                 }
             } else {
                 err = GetLastError();
-                fprintf(stderr, "%s: IOCTL_SCSI_GET_INQUIRY_DATA failed "
-                        "err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
+                pr2serr("%s: IOCTL_SCSI_GET_INQUIRY_DATA failed err=%lu\n"
+                        "\t%s", adapter_name, err,
+                        get_err_str(err, sizeof(b), b));
             }
             CloseHandle(fh);
         } else {
             if (verbose > 3) {
                 err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
+                pr2serr("%s: CreateFile failed err=%lu\n\t%s", adapter_name,
+                        err, get_err_str(err, sizeof(b), b));
             }
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
@@ -400,7 +422,7 @@ enum_volumes(char letter)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        pr2serr("%s: enter\n", __func__ );
     for (k = 0; k < 24; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name), "\\\\.\\%c:", 'C' + k);
@@ -410,12 +432,12 @@ enum_volumes(char letter)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n", __FUNCTION__ );
+                pr2serr("%s: query_dev_property failed\n", __func__ );
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n", __FUNCTION__ );
+                    pr2serr("%s: query_dev_uid failed\n", __func__ );
             } else
                 tmp_se.qp_uid_valid = 1;
             if (('\0' == letter) || (letter == tmp_se.name[0]))
@@ -438,7 +460,7 @@ enum_pds(void)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        pr2serr("%s: enter\n", __func__ );
     for (k = 0; k < MAX_PHYSICALDRIVE_NUM; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name),
@@ -449,12 +471,12 @@ enum_pds(void)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n", __FUNCTION__ );
+                pr2serr("%s: query_dev_property failed\n", __func__ );
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n", __FUNCTION__ );
+                    pr2serr("%s: query_dev_uid failed\n", __func__ );
             } else
                 tmp_se.qp_uid_valid = 1;
             hole_count = 0;
@@ -463,8 +485,8 @@ enum_pds(void)
         } else {
             if (verbose > 3) {
                 err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
+                pr2serr("%s: CreateFile failed err=%lu\n\t%s", adapter_name,
+                        err, get_err_str(err, sizeof(b), b));
             }
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
@@ -485,7 +507,7 @@ enum_cdroms(void)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        pr2serr("%s: enter\n", __func__ );
     for (k = 0; k < MAX_CDROM_NUM; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name), "\\\\.\\CDROM%d", k);
@@ -495,12 +517,12 @@ enum_cdroms(void)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n", __FUNCTION__ );
+                pr2serr("%s: query_dev_property failed\n", __func__ );
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n", __FUNCTION__ );
+                    pr2serr("%s: query_dev_uid failed\n", __func__ );
             } else
                 tmp_se.qp_uid_valid = 1;
             hole_count = 0;
@@ -509,8 +531,8 @@ enum_cdroms(void)
         } else {
             if (verbose > 3) {
                 err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
+                pr2serr("%s: CreateFile failed err=%lu\n\t%s", adapter_name,
+                        err, get_err_str(err, sizeof(b), b));
             }
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
@@ -531,7 +553,7 @@ enum_tapes(void)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        pr2serr("%s: enter\n", __func__ );
     for (k = 0; k < MAX_TAPE_NUM; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name), "\\\\.\\TAPE%d", k);
@@ -541,12 +563,12 @@ enum_tapes(void)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n", __FUNCTION__ );
+                pr2serr("%s: query_dev_property failed\n", __func__ );
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n", __FUNCTION__ );
+                    pr2serr("%s: query_dev_uid failed\n", __func__ );
             } else
                 tmp_se.qp_uid_valid = 1;
             hole_count = 0;
@@ -555,8 +577,8 @@ enum_tapes(void)
         } else {
             if (verbose > 3) {
                 err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
+                pr2serr("%s: CreateFile failed err=%lu\n\t%s", adapter_name,
+                        err, get_err_str(err, sizeof(b), b));
             }
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
@@ -605,7 +627,8 @@ do_wscan(char letter, int show_bt, int scsi_scan)
                 printf("[%4s+] ", sp->volume_letters);
             if (sp->qp_descriptor_valid) {
                 if (show_bt)
-                    printf("<%s>  ", get_bus_type(sp->qp_descriptor.desc.BusType));
+                    printf("<%s>  ",
+                           get_bus_type(sp->qp_descriptor.desc.BusType));
                 j = sp->qp_descriptor.desc.VendorIdOffset;
                 if (j > 0)
                     printf("%s  ", sp->qp_descriptor.raw + j);
@@ -648,7 +671,7 @@ sg_do_wscan(char letter, int do_scan, int verb)
         ret = do_wscan(letter, show_bt, scsi_scan);
         free(storage_arr);
     } else {
-        fprintf(stderr, "Failed to allocate storage_arr on heap\n");
+        pr2serr("Failed to allocate storage_arr on heap\n");
         ret = SG_LIB_SYNTAX_ERROR;
     }
     return ret;
